<?php

namespace Drupal\commerce_byjuno\Plugin\Commerce\CheckoutPane;

use Drupal\commerce\Response\NeedsRedirectException;
use Drupal\commerce_checkout\Plugin\Commerce\CheckoutPane\CheckoutPaneBase;
use Drupal\commerce_payment\Exception\DeclineException;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;

/**
 *
 * This pane should be added to the "review" step, and have a higher weight
 * than the "review" pane.
 *
 * @CommerceCheckoutPane(
 *   id = "commerce_byjuno_method",
 *   label = @Translation("Byjuno Invoice"),
 *   default_step = "review",
 *   wrapper_element = "fieldset",
 * )
 */
class ByjunoMethodPane extends CheckoutPaneBase {

  public function isVisible() {
    /** @var \Drupal\commerce_payment\Entity\PaymentMethodInterface $payment_method */
    $payment_method = $this->order->payment_method->entity;

    // Do nothing if this order does not involve Byjuno.
    if (!empty($payment_method) && $payment_method->bundle() == 'byjuno_invoice') {
      return TRUE;
    }

    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function buildPaneForm(array $pane_form, FormStateInterface $form_state, array &$complete_form) {
    // A payment gateway must be set before this pane is used.
    if ($this->order->get('payment_method')->isEmpty()) {
      \Drupal::logger('commerce_byjuno')->error('No payment gateway selected.');
      $this->redirectToPreviousStep();
    }

    /** @var \Drupal\commerce_payment\Entity\PaymentMethodInterface $payment_method */
    $payment_method = $this->order->payment_method->entity;

    /** @var \Drupal\commerce_byjuno\Plugin\Commerce\PaymentGateway\ByjunoInterface $payment_gateway */
    $payment_gateway_plugin = $payment_method->getPaymentGateway()->getPlugin();

    if (!$payment_method->getRemoteId()) {
      // Reserve the payment
      try {
        $payment_method = $payment_gateway_plugin->reservePayment($this->order);
      }
      catch (DeclineException $e) {
        \Drupal::logger('commerce_byjuno')->error($e->getMessage());

        // If this happens we are left with a payment method that doesn't
        // have a transaction ID set, and cannot be used. To avoid it showing up
        // in the list of possible existing payment methods, delete it.
        $this->order->set('payment_method', NULL);
        $this->order->save();
        $payment_method->delete();

        $this->redirectToPreviousStep();
      }
    }

    $pane_form['notice'] = [
      '#markup' => 'Credit approved',
    ];

    return $pane_form;
  }

  /**
   * Redirects to a previous checkout step on error.
   *
   * @throws \Drupal\commerce\Response\NeedsRedirectException
   */
  protected function redirectToPreviousStep() {
    $url = Url::fromRoute('commerce_checkout.form', [
      'commerce_order' => $this->order->id(),
      'step' => $this->checkoutFlow->getPreviousStepId($this->getStepId()),
    ], ['absolute' => TRUE]);
    throw new NeedsRedirectException($url->toString());
  }

}
